import types

HEADER_1_4S2M__MOTOR1 = 0x0011
HEADER_1_4S2M__MOTOR2 = 0x0012
HEADER_1_4S2M__MOTOR1_OPT = 0x0111
HEADER_1_4S2M__MOTOR2_OPT = 0x0112
HEADER_1_4S2M__MOTOR1_FAST = 0x0031
HEADER_1_4S2M__MOTOR2_FAST = 0x0032
HEADER_2_4S2M__MOTOR1 = 0x0021
HEADER_2_4S2M__MOTOR2 = 0x0022
HEADER_2_4S2M__MOTOR1_OPT = 0x0121
HEADER_2_4S2M__MOTOR2_OPT = 0x0122
HEADER_2_4S2M__MOTOR1_FAST = 0x0041
HEADER_2_4S2M__MOTOR2_FAST = 0x0042

MIN_STEP_COUNT = 0
MAX_STEP_COUNT0 = 255
MAX_STEP_COUNT1 = 0x7FFF

FIRMWARE_VER_REG = 0xFFFF

class CheckVer:
    _ACK_TOUT_SEC = 0.25

    def __init__(self, modbus_transport):
        self._modbus_transport = modbus_transport

    def check_ver(self, res):
        self._modbus_transport.read_reg(FIRMWARE_VER_REG, res, CheckVer._ACK_TOUT_SEC)


class MotorControl:
    _CLOCKWISE = 0x00
    _ANTICLOCKWISE = 0x100
    _ACK_TOUT_SEC_MIN = 0.25

    def __init__(self, modbus_transport, motor_address, v=0):
        if v > 1:
            if v > 18:
                if motor_address not in [HEADER_1_4S2M__MOTOR1, HEADER_1_4S2M__MOTOR2,
                                     HEADER_2_4S2M__MOTOR1, HEADER_2_4S2M__MOTOR2,
                                     HEADER_1_4S2M__MOTOR1_OPT, HEADER_1_4S2M__MOTOR2_OPT,
                                     HEADER_2_4S2M__MOTOR1_OPT, HEADER_2_4S2M__MOTOR2_OPT,
                                     HEADER_1_4S2M__MOTOR1_FAST, HEADER_1_4S2M__MOTOR2_FAST,
                                     HEADER_2_4S2M__MOTOR1_FAST, HEADER_2_4S2M__MOTOR2_FAST]:
                    raise ValueError("Invalid motor_address")
            else:
                if motor_address not in [HEADER_1_4S2M__MOTOR1, HEADER_1_4S2M__MOTOR2,
                                     HEADER_2_4S2M__MOTOR1, HEADER_2_4S2M__MOTOR2,
                                     HEADER_1_4S2M__MOTOR1_FAST, HEADER_1_4S2M__MOTOR2_FAST,
                                     HEADER_2_4S2M__MOTOR1_FAST, HEADER_2_4S2M__MOTOR2_FAST]:
                    raise ValueError("Invalid motor_address")
        else:
            if motor_address not in [HEADER_1_4S2M__MOTOR1, HEADER_1_4S2M__MOTOR2,
                                     HEADER_2_4S2M__MOTOR1, HEADER_2_4S2M__MOTOR2]:
                raise ValueError("Invalid motor_address")

        self._modbus_transport = modbus_transport
        self._motor_address = motor_address

    @staticmethod
    def _validate_step_count(step_count, v):
        if v > 1:
            return MIN_STEP_COUNT <= step_count <= MAX_STEP_COUNT1
        else:
            return MIN_STEP_COUNT <= step_count <= MAX_STEP_COUNT0

    def rotate_cw(self, step_count, v=0, speed_count=0):
        """
        :type step_count: int
        :rtype: int
        """
        if not MotorControl._validate_step_count(step_count, v):
            print "the specified steps number is not supported by fw version ", v
            return types.RC_FAIL

        if v > 18:
            if speed_count > 63:
                print "the specified speed value is too big for fw version ", v
                return types.RC_FAIL
            if self._motor_address in [HEADER_1_4S2M__MOTOR1_OPT, HEADER_1_4S2M__MOTOR2_OPT,
                                       HEADER_2_4S2M__MOTOR1_OPT, HEADER_2_4S2M__MOTOR2_OPT]:
                tout = 2 + 2 * step_count * 0.003 * (speed_count + 1)
            else:
                tout = MotorControl._ACK_TOUT_SEC_MIN + step_count * 0.003 * (speed_count + 1)
        else:
            if self._motor_address in [HEADER_1_4S2M__MOTOR1_FAST, HEADER_1_4S2M__MOTOR2_FAST,
                                       HEADER_2_4S2M__MOTOR1_FAST, HEADER_2_4S2M__MOTOR2_FAST]:
                tout = MotorControl._ACK_TOUT_SEC_MIN + step_count * 0.003
            else:
                tout = MotorControl._ACK_TOUT_SEC_MIN + 3 * step_count * 0.003 * 2

        if v > 1:
            if v > 18:
                if step_count > 255:
                    powerVal = 0
                    stepVal = int(step_count / 2)
                    while stepVal > 0:
                        powerVal += 1
                        stepVal = int(stepVal / 2)
                    motorVal = MotorControl._CLOCKWISE + 0x8000 + powerVal
                else:
                    motorVal = MotorControl._CLOCKWISE + step_count
                motorVal += speed_count * 0x200
            else:
                motorVal = MotorControl._CLOCKWISE + int(int(step_count / 256) * 0x200) + (step_count % 256)
        else:
            motorVal = MotorControl._CLOCKWISE + (step_count % 256)

        retcode = self._modbus_transport.write_reg(self._motor_address, motorVal, tout)
        return retcode

    def rotate_cw_get_reply(self, step_count, res, v=0, speed_count=0):
        """
        :type step_count: int
        :rtype: int
        """
        if not MotorControl._validate_step_count(step_count, v):
            print "the specified steps number is not supported by fw version ", v
            res.append(types.RC_FAIL)
            return types.RC_FAIL

        if v > 18:
            if speed_count > 63:
                print "the specified speed value is too big for fw version ", v
                return types.RC_FAIL
            if self._motor_address in [HEADER_1_4S2M__MOTOR1_OPT, HEADER_1_4S2M__MOTOR2_OPT,
                                       HEADER_2_4S2M__MOTOR1_OPT, HEADER_2_4S2M__MOTOR2_OPT]:
                tout = 2 + 2 * step_count * 0.003 * (speed_count + 1)
            else:
                tout = MotorControl._ACK_TOUT_SEC_MIN + step_count * 0.003 * (speed_count + 1)
        else:
            if self._motor_address in [HEADER_1_4S2M__MOTOR1_FAST, HEADER_1_4S2M__MOTOR2_FAST,
                                       HEADER_2_4S2M__MOTOR1_FAST, HEADER_2_4S2M__MOTOR2_FAST]:
                tout = MotorControl._ACK_TOUT_SEC_MIN + step_count * 0.003
            else:
                tout = MotorControl._ACK_TOUT_SEC_MIN + 3 * step_count * 0.003 * 2

        if v > 1:
            if v > 18:
                if step_count > 255:
                    powerVal = 0
                    stepVal = int(step_count / 2)
                    while stepVal > 0:
                        powerVal += 1
                        stepVal = int(stepVal / 2)
                    motorVal = MotorControl._CLOCKWISE + 0x8000 + powerVal
                else:
                    motorVal = MotorControl._CLOCKWISE + step_count
                motorVal += speed_count * 0x200
            else:
                motorVal = MotorControl._CLOCKWISE + int(int(step_count / 256) * 0x200) + (step_count % 256)
        else:
            motorVal = MotorControl._CLOCKWISE + (step_count % 256)

        retcode = self._modbus_transport.write_reg_get_reply(self._motor_address, motorVal, tout, res)
        return retcode

    def rotate_ccw(self, step_count, v=0, speed_count=0):
        """
        :type step_count: int
        :rtype: int
        """
        if not MotorControl._validate_step_count(step_count, v):
            return types.RC_FAIL

        if v > 18:
            if speed_count > 63:
                print "the specified speed value is too big for fw version ", v
                return types.RC_FAIL
            if self._motor_address in [HEADER_1_4S2M__MOTOR1_OPT, HEADER_1_4S2M__MOTOR2_OPT,
                                       HEADER_2_4S2M__MOTOR1_OPT, HEADER_2_4S2M__MOTOR2_OPT]:
                tout = 2 + 2 * step_count * 0.003 * (speed_count + 1)
            else:
                tout = MotorControl._ACK_TOUT_SEC_MIN + step_count * 0.003 * (speed_count + 1)
        else:
            if self._motor_address in [HEADER_1_4S2M__MOTOR1_FAST, HEADER_1_4S2M__MOTOR2_FAST,
                                       HEADER_2_4S2M__MOTOR1_FAST, HEADER_2_4S2M__MOTOR2_FAST]:
                tout = MotorControl._ACK_TOUT_SEC_MIN + step_count * 0.003
            else:
                tout = MotorControl._ACK_TOUT_SEC_MIN + 3 * step_count * 0.003 * 2

        if v > 1:
            if v > 18:
                if step_count > 255:
                    powerVal = 0
                    stepVal = int(step_count / 2)
                    while stepVal > 0:
                        powerVal += 1
                        stepVal = int(stepVal / 2)
                    motorVal = MotorControl._ANTICLOCKWISE + 0x8000 + powerVal
                else:
                    motorVal = MotorControl._ANTICLOCKWISE + step_count
                motorVal += speed_count * 0x200
            else:
                motorVal = MotorControl._ANTICLOCKWISE + int(int(step_count / 256) * 0x200) + (step_count % 256)
        else:
            motorVal = MotorControl._ANTICLOCKWISE + (step_count % 256)

        retcode = self._modbus_transport.write_reg(self._motor_address, motorVal, tout)
        return retcode

    def rotate_ccw_get_reply(self, step_count, res, v=0, speed_count=0):
        """
        :type step_count: int
        :rtype: int
        """
        if not MotorControl._validate_step_count(step_count, v):
            res.append(types.RC_FAIL)
            return types.RC_FAIL

        if v > 18:
            if speed_count > 63:
                print "the specified speed value is too big for fw version ", v
                return types.RC_FAIL
            if self._motor_address in [HEADER_1_4S2M__MOTOR1_OPT, HEADER_1_4S2M__MOTOR2_OPT,
                                       HEADER_2_4S2M__MOTOR1_OPT, HEADER_2_4S2M__MOTOR2_OPT]:
                tout = 2 + 2 * step_count * 0.003 * (speed_count + 1)
            else:
                tout = MotorControl._ACK_TOUT_SEC_MIN + step_count * 0.003 * (speed_count + 1)
        else:
            if self._motor_address in [HEADER_1_4S2M__MOTOR1_FAST, HEADER_1_4S2M__MOTOR2_FAST,
                                       HEADER_2_4S2M__MOTOR1_FAST, HEADER_2_4S2M__MOTOR2_FAST]:
                tout = MotorControl._ACK_TOUT_SEC_MIN + step_count * 0.003
            else:
                tout = MotorControl._ACK_TOUT_SEC_MIN + 3 * step_count * 0.003 * 2

        if v > 1:
            if v > 18:
                if step_count > 255:
                    powerVal = 0
                    stepVal = int(step_count / 2)
                    while stepVal > 0:
                        powerVal += 1
                        stepVal = int(stepVal / 2)
                    motorVal = MotorControl._ANTICLOCKWISE + 0x8000 + powerVal
                else:
                    motorVal = MotorControl._ANTICLOCKWISE + step_count
                motorVal += speed_count * 0x200
            else:
                motorVal = MotorControl._ANTICLOCKWISE + int(int(step_count / 256) * 0x200) + (step_count % 256)
        else:
            motorVal = MotorControl._ANTICLOCKWISE + (step_count % 256)

        retcode = self._modbus_transport.write_reg_get_reply(self._motor_address, motorVal, tout, res)
        return retcode
